// compression utility for 4k intros
// features lzss77 compression with variable length codes bitstream
// reverse engineered mesha/rhyme packer
// done by shodan@chat.ru

// there are three types of lz-codes possible in the bitstream:
//   00 xxxx                  - len=1               distance=xxxx+1
//   01111111 lzcode1 lzcode2 - len=lzcode1         distance=lzcode2
//   01..1 lzcode             - len=(numer of 1's)  distance=lzcode
// original version:
// each lzcode is either 5bit or 9bit, plus one bit in the beginning
// indicating how long (5bit/9bit) that lzcode is. 5bit lzcodes code
// values 0x01..0x20, 9bit lzcodes code values 0x21..0x1ff+0x21, and
// thus the window size equals max match len and equals 0x220.
// upgraded version:
// each lzcode is either s1-bit or s2-bit.

#define S1          6
#define S2          11
#define MAX1        (1 << S1)
#define MAX2        (1 << S2)
#define WINSIZE     (MAX1 + MAX2)
#define MAXMATCHLEN WINSIZE
#define MAXDISTANCE WINSIZE

#include <stdio.h>

char source[8192], packed[8192], stub[256];
int spos = 0, ppos = 0, pbit = 0, srclen, stublen;
FILE *dbg;

int matchlen(int p1, int p2) {
  int i = 0;
  while (source[p1 + i] == source[p2 + i] && p2 + i < srclen &&
    p1 + i < srclen && i < MAXMATCHLEN) i++;
  return i;
}

void putbits(int bits, int value) {
  int b;

  while (bits--) {
    if (pbit == 8) { pbit = 0; ppos++; }
    b = (value >> bits) & 0x01;
    packed[ppos] |= (b << pbit);
    pbit++;
  }
}

int lzcodelen(int code) {
  if (code <= MAX1) return 1 + S1; else return 1 + S2;
}

int cost (int len, int distance) {
  if (len == 0) return 255;
  if (len == 1) {
    if (distance < 0x11) return 6; else return 255;
  }
  if (len < 8) return (lzcodelen(distance) + len + 1);
  return (8 + lzcodelen(len - 1) + lzcodelen(distance));
}

int putlzcode(int code) {
  code--;
  if (code < MAX1) {
    putbits(1, 0);
    putbits(S1, code);
    return 1 + S1;
  } else {
    putbits(1, 1);
    putbits(S2, code - MAX1);
    return 1 + S2;
  }
}

int putcode (int len, int distance) {
  int t;

  if (len == 1 && distance < 0x11) {
    putbits(2, 2);
    putbits(4, distance - 1);
    return 6;
  }
  if (len < 8) {
    putbits(1, 1);
    putbits(len - 1, 0xFF);
    putbits(1, 0);
    t = putlzcode(distance);
    return (1 + len + t);
  }
  putbits(8, 0xFF);
  t = 8;
  t += putlzcode(len - 1);
  t += putlzcode(distance);
  return t;
}

void putld(int len, int distance) {
  if (distance == 0) {
    //fprintf(dbg, " %x | %x | copy %x\n", spos, ppos, source[spos]);
    putbits(1, 0);
    putbits(8, source[spos]);
    spos++;
  } else {
    //fprintf(dbg, " %x | %x | %d %d\n", spos, ppos, len, distance);
    putcode(len, distance);
    spos += len;
  }
}

void bestmatch(int pos, int *len, int *distance) {
  int i, p1, cl, cc, tl, tc;

  p1 = (pos <= MAXDISTANCE) ? 0 : pos - MAXDISTANCE;
  cl = 1;
  cc = 9;
  *distance = 0;
  *len = cl;
  for (i = p1; i < pos; i++) {
    tl = matchlen(i, pos);
    tc = cost(tl, pos - i);
    if (cc * tl >= cl * tc) {
      cc = tc;
      cl = tl;
      *distance = pos - i;
      *len = cl;
    }
  }
}

int sign(int a) {
  if (a == 0) return 0;
  if (a < 0) return -1;
  return 1;
}

void main(int argc, char *argv[]) {
  FILE *f;
  int i, j, distance, len;

  int l[WINSIZE+8], d[WINSIZE+8], c[WINSIZE+8], n;
  int b, bl, bl0, bc;
  int p1, cl, cc, tl, tc, td;
  float c1, c2;

  //dbg = fopen("debug", "wt+");

  if ((f = fopen("stub2.com", "rb")) == NULL) {
    printf("error: unable to open stub2.com\n");
    exit(0);
  }
  stublen = fread(&stub, 1, sizeof stub, f);

  fclose(f);

  if (argc != 3) {
    printf("meshpack[.exe] <source.com> <dest.com>\n");
    exit(0);
  }
  memset(packed, 0x00, sizeof packed);

  if ((f = fopen(argv[1], "rb")) == NULL) {
    printf("error: unable to open %s\n", argv[1]);
    exit(0);
  }
  //fseek(f, 0x1000, SEEK_SET); // unpacked code will execute from 1000h
  srclen = fread(&source, 1, sizeof source, f);
  fclose(f);

  spos = ppos = 0;
  while (spos < srclen) {

      p1 = (spos <= MAXDISTANCE) ? 0 : spos - MAXDISTANCE;
      n = 0;
      l[0] = 1;
      c[0] = 9;
      d[0] = 0;
      for (i = p1; i < spos; i++) {
        n++;
        l[n] = matchlen(i, spos);
        c[n] = cost(l[n], spos - i);
        d[n] = spos - i;
      }

      b = 0;
      bestmatch(spos + l[0], &tl, &td);
      bc = c[0] + cost(tl, td);
      bl = l[0] + tl;
      bl0 = l[0];
      for (i = 1; i <= n; i++) {
        if (l[i] != 0) {
	  for (j = 1; j <= l[i]; j++) {
            bestmatch(spos + j, &tl, &td);
            tc = cost(tl, td) + cost(j, d[i]);
            tl += j;

	    // this is quite weird way to find out whether one lz-sequence
	    // is better than other, in fact i'm not even sure why it does
	    // work. but it works and shows very good results and hey i am
	    // tired of messing with this already.. hint: sign(bl - tl) is
	    // way too big bonus, the compare must be done cleaner, for it
	    // yields (tiny) bit leaks now.

	    c1 = (float)bc / (float)bl;
	    c2 = (float)tc / (float)tl;
            if (c1 >= sign(bl - tl) + c2) {
              b = i;
              bc = tc;
              bl = tl;
	      bl0 = j;
            }
	  }
        }
      }

      len = bl0;
      distance = d[b];
      putld(len, distance);

    printf("%d/%d\r", spos, srclen);
    fflush(stdout);
  }

  if ((f = fopen(argv[2], "wb+")) == NULL) {
    printf("error: unable to open %s\n", argv[1]);
    exit(0);
  }
  fwrite(&stub, 1, stublen, f);
  fwrite(&packed, 1, ppos + 1, f);
  fclose(f);
  //fclose(dbg);

  printf("compressed size=%d   original size=%d\n", ppos + stublen + 1, srclen);
}